import os


def get_proxy(proxy: str = None):
    proxy = (
            proxy
            or os.environ.get("all_proxy")
            or os.environ.get("ALL_PROXY")
            or os.environ.get("https_proxy")
            or os.environ.get("HTTPS_PROXY")
            or None
    )
    if proxy is not None and proxy.startswith("socks5h://"):
        proxy = "socks5://" + proxy[len("socks5h://"):]
    return proxy



import asyncio
import json
import re
import secrets

import aiohttp

from re_edge_gpt.chat.constants import SUNO_HEADER


async def generate_suno_music(cookies: list[dict], iframe_id: str, request_id: str):
    new_cookies = {}
    for cookie in cookies:
        new_cookies[cookie["name"]] = cookie["value"]
    async with aiohttp.ClientSession(
            headers=SUNO_HEADER, cookies=new_cookies
    ) as session:
        generate_url = f"https://www.bing.com/videos/music?vdpp=suno&kseed=7500&SFX=2&q=&" \
                       f"iframeid={iframe_id}&requestid={request_id}"
        response = await session.get(generate_url)
        if response.status != 200:
            raise ConnectionError("Generate suno music failed")
        skey = re.findall('skey=(.*?)&amp;', await response.text())
        if skey:
            skey = skey[0]
        else:
            raise ConnectionError("Generate suno music failed cause by: skey")
        real_generate_url = f"https://www.bing.com/videos/api/custom/music?skey={skey}" \
                            f"&safesearch=Moderate&vdpp=suno&requestid={request_id}" \
                            f"&ig={secrets.token_hex(32).upper()}&iid=vsn&sfx=1"
        async with aiohttp.ClientSession(
                headers={"Referer": real_generate_url}, cookies=new_cookies
        ) as real_session:
            while True:
                await asyncio.sleep(3)
                response = await real_session.get(real_generate_url)
                suno_response: dict = json.loads(await response.text())
                suno_response = {"RawResponse": json.loads(suno_response.get("RawResponse"))}
                if suno_response.get("RawResponse").get("status") == "running":
                    continue
                if suno_response.get("RawResponse").get("status") != "complete":
                    raise ConnectionError("Generate suno music failed cause by: status error")
                suno_response = suno_response.get("RawResponse")
                return {
                    "Image": f"https://th.bing.com/th?&id={suno_response.get('imageKey')}",
                    "Audio": f"https://th.bing.com/th?&id={suno_response.get('audioKey')}",
                    "Video": f"https://th.bing.com/th?&id={suno_response.get('videoKey')}",
                    "Duration": f"https://th.bing.com/th?&id={suno_response.get('duration')}",
                    "Style": f"https://th.bing.com/th?&id={suno_response.get('musicalStyle')}",
                    "Title": f"https://th.bing.com/th?&id={suno_response.get('gptPrompt')}",
                    "Lyrics": f"https://th.bing.com/th?&id={suno_response.get('lyrics')}",
                }
            raise ConnectionError("Generate suno music failed")



import socket
import uuid

take_ip_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
take_ip_socket.connect(("8.8.8.8", 80))
FORWARDED_IP: str = take_ip_socket.getsockname()[0]
take_ip_socket.close()


DELIMITER = "\x1e"

PLUGINS = {
    "notebook": "c310c353-b9f0-4d76-ab0d-1dd5e979cf68",
    "instacart": "46664d33-1591-4ce8-b3fb-ba1022b66c11",
    "kayak": "d6be744c-2bd9-432f-95b7-76e103946e34",
    "klarna": "5f143ea3-8c80-4efd-9515-185e83b7cf8a",
    "opentable": "543a7b1b-ebc6-46f4-be76-00c202990a1b",
    "shop": "39e3566a-d481-4d99-82b2-6d739b1e716e",
    "suno": "22b7f79d-8ea4-437e-b5fd-3e21f09f7bc1"
}

HEADERS = {
    "accept": "application/json",
    "accept-language": "en;q=0.9,en-US;q=0.8",
    "accept-encoding": "gzip, deflate, br, zsdch",
    "content-type": "application/json",
    "sec-ch-ua": '"Not/A)Brand";v="8", "Chromium";v="126", "Microsoft Edge";v="126"',
    "sec-ch-ua-arch": '"x86"',
    "sec-ch-ua-bitness": '"64"',
    "sec-ch-ua-full-version": '"126.0.2592.68"',
    "sec-ch-ua-full-version-list": '"Not/A)Brand";v="8.0.0.0", "Chromium";v="126.0.6478.114", "Microsoft Edge";v="126.0.2592.68"',
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-model": "",
    "sec-ch-ua-platform": '"Windows"',
    "sec-ch-ua-platform-version": '"15.0.0"',
    "sec-fetch-dest": "empty",
    "sec-fetch-mode": "cors",
    "sec-fetch-site": "same-origin",
    "sec-ms-gec-version": "1-126.0.2592.68",
    "x-ms-client-request-id": str(uuid.uuid4()),
    "x-ms-useragent": "azsdk-js-api-client-factory/1.0.0-beta.1 core-rest-pipeline/1.15.1 OS/Windows",
    "Referer": "https://www.bing.com/search?form=NTPCHB&q=Bing+AI&showconv=1",
    "Referrer-Policy": "origin-when-cross-origin",
    "x-forwarded-for": FORWARDED_IP,
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36 Edg/126.0.0.0",
}

HEADERS_INIT_CONVER = {
    "accept": "application/json",
    "accept-language": "en;q=0.9,en-US;q=0.8",
    "cache-control": "max-age=0",
    "sec-ch-ua": '"Not/A)Brand";v="8", "Chromium";v="126", "Microsoft Edge";v="126"',
    "sec-ch-ua-arch": '"x86"',
    "sec-ch-ua-bitness": '"64"',
    "sec-ch-ua-full-version": '"126.0.2592.68"',
    "sec-ch-ua-full-version-list": '"Not/A)Brand";v="8.0.0.0", "Chromium";v="126.0.6478.114", "Microsoft Edge";v="126.0.2592.68"',
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-model": '""',
    "sec-ch-ua-platform": '"Windows"',
    "sec-ch-ua-platform-version": '"15.0.0"',
    "upgrade-insecure-requests": "1",
    "x-edge-shopping-flag": "1",
    "X-Ms-Useragent": "azsdk-js-api-client-factory/1.0.0-beta.1 core-rest-pipeline/1.15.1 OS/Windows",
    "x-forwarded-for": FORWARDED_IP,
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36 Edg/126.0.0.0",
}

IMAGE_HEADER = {
    "Accept": "*/*",
    "Accept-Encoding": "gzip, deflate, br",
    "Content-Type": "multipart/form-data",
    "Referer": "https://www.bing.com/search?q=Bing+AI&showconv=1",
    "sec-ch-ua": '"Not/A)Brand";v="8", "Chromium";v="126", "Microsoft Edge";v="126"',
    "sec-ch-ua-arch": '"x86"',
    "sec-ch-ua-bitness": '"64"',
    "sec-ch-ua-full-version": '"126.0.2592.68"',
    "sec-ch-ua-full-version-list": '"Not/A)Brand";v="8.0.0.0", "Chromium";v="126.0.6478.114", "Microsoft Edge";v="126.0.2592.68"',
    "Sec-Ch-Ua-Mobile": "?0",
    "Sec-Ch-Ua-Platform": "Windows",
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "same-origin",
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36 Edg/126.0.0.0",
}

SYDNEY_INIT_HEADER = HEADERS_INIT_CONVER.update(
    {
        "Referer": "https://copilot.microsoft.com/",
        "X-Edge-Shopping-Flag": "0",
    }
)

SYDNEY_HEADER = HEADERS.update(
    {
        "Host": "sydney.bing.com",
        "Cache-Control": "no-cache",
        "Connection": "Upgrade",
        "Origin": "https://www.bing.com",
        "Pragma": "no-cache",
    }
)

SUNO_HEADER = {
  "Referer": "https://www.bing.com/search?q=Bing+AI&showconv=1&wlexpsignin=1"
}

BUNDLE_VERSION = "1.1782.0"
APP_ID = "6c0f12ef-97d3-4869-bc42-c1d9bdb4a759"



from typing import Union

import pyttsx3
from PySide6.QtCore import QTimer, Qt
from PySide6.QtWidgets import QBoxLayout, QWidget, QPushButton, QHBoxLayout, QTextEdit, QMessageBox

from re_edge_gpt.ui.chat.toast import ChatToast
from re_edge_gpt.ui.chat.chat_thread import MESSAGE_QUEUE, EXCEPTION_QUEUE


class ChatInputDialog(QWidget):
    def __init__(self, close_time: int = 10000, font_size: int = 16, show_toast: bool = True):
        super().__init__()
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.box_layout = QBoxLayout(QBoxLayout.Direction.TopToBottom)
        self.chat_input = QTextEdit()
        self.send_text_button = QPushButton()
        self.send_text_button.setText("Send prompt")
        self.box_h_layout = QHBoxLayout()
        self.box_h_layout.addWidget(self.send_text_button)
        self.box_layout.addWidget(self.chat_input)
        self.box_layout.addLayout(self.box_h_layout)
        self.setWindowTitle("Please input prompt")
        self.setLayout(self.box_layout)
        # Show toast?
        self.show_toast = show_toast
        # Get message timer
        self.get_message_timer = QTimer()
        self.get_message_timer.setInterval(1000)
        self.get_message_timer.timeout.connect(self.get_message)
        self.get_message_timer.start()
        # Check error timer
        self.check_error_timer = QTimer()
        self.check_error_timer.setInterval(1000)
        self.check_error_timer.timeout.connect(self.check_error)
        self.check_error_timer.start()
        # Toast
        self.close_time = close_time
        self.font_size = font_size
        self.toast_widget: Union[ChatToast, None] = None
        # Text to speech
        self.engine = pyttsx3.init()

    def get_message(self):
        if not MESSAGE_QUEUE.empty():
            text = MESSAGE_QUEUE.get_nowait()
            if self.show_toast:
                self.toast_widget = ChatToast(
                    text=text, close_time=self.close_time, font_size=self.font_size)
                self.toast_widget.showFullScreen()
            # self.engine.say(text)

    def show_text(self):
        if self.toast_widget is not None and self.toast_widget.isVisible() is False:
            self.get_message()
        elif self.toast_widget is None:
            self.get_message()

    def check_error(self):
        if not EXCEPTION_QUEUE.empty():
            gpt_error_messagebox = QMessageBox(self)
            gpt_error_messagebox.setText(
                "Catch exception" + "\n"
                + EXCEPTION_QUEUE.get_nowait()
            )
            gpt_error_messagebox.show()

    def close(self) -> bool:
        self.get_message_timer.stop()
        self.check_error_timer.stop()
        self.destroy()
        return False


import argparse
import asyncio
import contextlib
import json
import os
import random
import sys
import time
from typing import Dict
from typing import List
from typing import Union

import httpx
import regex
import requests
from requests.utils import requote_uri

from re_edge_gpt.chat.proxy import get_proxy
from re_edge_gpt.utils.exception.exception_message import sending_message, error_being_reviewed_prompt, \
    error_blocked_prompt, \
    error_unsupported_lang, error_timeout, error_noresults, error_no_images, download_message, error_image_create_failed
from re_edge_gpt.utils.exception.exceptions import UnSupportLanguage, PromptBlocked, ImageCreateFailed, \
    NoResultsFound, AuthCookieError, LimitExceeded, InappropriateContentType, ResponseError

FORWARDED_IP = f"1.0.0.{random.randint(0, 255)}"

BING_URL = os.getenv("BING_URL", "https://www.bing.com")

HEADERS = {
    "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,"
              "image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.77",
    "accept-language": "en;q=0.9,en-US;q=0.8",
    "cache-control": "max-age=0",
    "content-type": "application/x-www-form-urlencoded",
    "referrer": "https://www.bing.com/images/create/",
    "origin": "https://copilot.microsoft.com/",
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                  "AppleWebKit/537.36 (KHTML, like Gecko) "
                  "Chrome/120.0.0.0 "
                  "Safari/537.36 "
                  "Edg/120.0.2210.91",
}


def check_response(response):
    # check for content waring message
    if "this prompt is being reviewed" in response.text.lower():
        raise UnSupportLanguage(
            error_being_reviewed_prompt,
        )
    if "this prompt has been blocked" in response.text.lower():
        raise PromptBlocked(
            error_blocked_prompt,
        )
    if (
            "we're working hard to offer image creator in more languages"
            in response.text.lower()
    ):
        raise UnSupportLanguage(error_unsupported_lang)


class ImageGen:
    """
    Image generation by Microsoft Bing
    Parameters:
        auth_cookie: str
    Optional Parameters:
        debug_file: str
        quiet: bool
        all_cookies: List[Dict]
    """

    def __init__(
            self,
            auth_cookie: str,
            quiet: bool = False,
            all_cookies: List[Dict] = None,
            proxy: str = None,
            proxy_user: Dict[str, str] = None
    ) -> None:
        if proxy_user is None:
            proxy_user = {"http_user": "http", "https_user": "https"}
        self.session: requests.Session = requests.Session()
        self.proxy: str = get_proxy(proxy)
        if self.proxy is not None:
            self.session.proxies.update({
                proxy_user.get("http_user", "http"): self.proxy,
                proxy_user.get("https_user", "https"): self.proxy
            })
        self.session.headers = HEADERS
        self.session.cookies.set("_U", auth_cookie)
        if all_cookies:
            for cookie in all_cookies:
                self.session.cookies.set(cookie["name"], cookie["value"])
        self.quiet = quiet

    def get_images(self, prompt: str, timeout: int = 200, max_generate_time_sec: int = 60) -> Union[list, None]:
        """
        Fetches image links from Bing
        Parameters:
            :param prompt: str -> prompt to gen image
            :param timeout: int -> timeout
            :param max_generate_time_sec: time limit of generate image
        """
        if not self.quiet:
            print(sending_message)
        url_encoded_prompt = requote_uri(prompt)
        payload = f"q={url_encoded_prompt}&qs=ds"
        url = f"{BING_URL}/images/create?q={url_encoded_prompt}&rt=4&FORM=GUH2CR"
        response = self.session.post(
            url,
            allow_redirects=False,
            data=payload,
            timeout=timeout,
        )
        check_response(response)
        if response.status_code != 302:
            # if rt4 fails, try rt3
            url = f"{BING_URL}/images/create?q={url_encoded_prompt}&rt=3&FORM=GUH2CR"
            response = self.session.post(url, allow_redirects=False, timeout=timeout)
            if response.status_code != 302:
                raise ImageCreateFailed(error_image_create_failed)
                # Get redirect URL
        redirect_url = response.headers["Location"].replace("&nfy=1", "")
        request_id = redirect_url.split("id=")[-1]
        self.session.get(f"{BING_URL}{redirect_url}")
        # https://www.bing.com/images/create/async/results/{ID}?q={PROMPT}
        polling_url = f"{BING_URL}/images/create/async/results/{request_id}?q={url_encoded_prompt}"
        # Poll for results
        if not self.quiet:
            print("Waiting for results...")
        start_wait = time.time()
        time_sec = 0
        while True:
            if int(time.time() - start_wait) > 200:
                raise TimeoutError(error_timeout)
            if not self.quiet:
                print(".", end="", flush=True)
            response = self.session.get(polling_url)
            if response.status_code != 200:
                raise NoResultsFound(error_noresults)
            if not response.text or response.text.find("errorMessage") != -1:
                time.sleep(1)
                time_sec = time_sec + 1
                if time_sec >= max_generate_time_sec:
                    raise TimeoutError("Out of generate time")
            else:
                break
        # Use regex to search for src=""
        image_links = regex.findall(r'src="([^"]+)"', response.text)
        # Remove size limit
        normal_image_links = [link.split("?w=")[0] for link in image_links]
        # Remove duplicates
        normal_image_links = list(set(normal_image_links))

        # Bad images
        bad_images = [
            "https://r.bing.com/rp/in-2zU3AJUdkgFe7ZKv19yPBHVs.png",
            "https://r.bing.com/rp/TX9QuO3WzcCJz1uaaSwQAz39Kb0.jpg",
        ]
        for img in normal_image_links:
            if img in bad_images:
                raise NoResultsFound("Bad images")
        # No images
        if not normal_image_links:
            raise NoResultsFound(error_no_images)
        return normal_image_links

    def save_images(
            self,
            links: list,
            output_dir: str,
            file_name: str = None,
            download_count: int = None,
    ) -> None:
        """
        Saves images to output directory
        Parameters:
            links: list[str]
            output_dir: str
            file_name: str
            download_count: int
        """
        if not self.quiet:
            print(download_message)
        with contextlib.suppress(FileExistsError):
            os.mkdir(output_dir)
        try:
            fn = f"{file_name}_" if file_name else ""
            jpeg_index = 0

            if download_count:
                links = links[:download_count]

            for link in links:
                while os.path.exists(
                        os.path.join(output_dir, f"{fn}{jpeg_index}.jpeg")
                ):
                    jpeg_index += 1
                response = self.session.get(link)
                if response.status_code != 200:
                    raise ResponseError(f"Could not download image response code {response.status_code}")
                # save response to file
                with open(
                        os.path.join(output_dir, f"{fn}{jpeg_index}.jpeg"), "wb"
                ) as output_file:
                    output_file.write(response.content)
                jpeg_index += 1

        except requests.exceptions.MissingSchema as url_exception:
            raise InappropriateContentType(
                "Inappropriate contents found in the generated images. Please try again or try another prompt.",
            ) from url_exception


class ImageGenAsync:
    """
    Image generation by Microsoft Bing
    Parameters:
        auth_cookie: str
    Optional Parameters:
        debug_file: str
        quiet: bool
        all_cookies: list[dict]
    """

    def __init__(
            self,
            auth_cookie: str = None,
            quiet: bool = False,
            all_cookies: List[Dict] = None,
            proxy: str = None
    ) -> None:
        if auth_cookie is None and not all_cookies:
            raise AuthCookieError("No auth cookie provided")
        self.proxy: str = get_proxy(proxy)
        self.session = httpx.AsyncClient(
            proxies=self.proxy,
            headers=HEADERS,
            trust_env=True,
        )
        if auth_cookie:
            self.session.cookies.update({"_U": auth_cookie})
        if all_cookies:
            for cookie in all_cookies:
                self.session.cookies.update(
                    {cookie["name"]: cookie["value"]},
                )
        self.quiet = quiet

    async def __aenter__(self):
        return self

    async def __aexit__(self, *excinfo) -> None:
        await self.session.aclose()

    async def get_images(self, prompt: str, timeout: int = 200, max_generate_time_sec: int = 60) -> Union[list, None]:
        """
        Fetches image links from Bing
        Parameters:
            :param prompt: str -> prompt to gen image
            :param timeout: int -> timeout
            :param max_generate_time_sec: time limit of generate image
        """
        if not self.quiet:
            print("Sending request...")
        url_encoded_prompt = requote_uri(prompt)
        url = f"{BING_URL}/images/create?q={url_encoded_prompt}&rt=4&FORM=GUH2CR"
        response = await self.session.post(
            url,
            follow_redirects=False,
            data={"q": url_encoded_prompt, "qs": "ds"},
            timeout=timeout
        )
        check_response(response)
        if response.status_code != 302:
            url = f"{BING_URL}/images/create?q={url_encoded_prompt}&rt=3&FORM=GUH2CR"
            response = await self.session.post(
                url,
                follow_redirects=False,
                timeout=timeout,
            )
        if response.status_code != 302:
            raise ImageCreateFailed(error_image_create_failed)
        # Get redirect URL
        redirect_url = response.headers["Location"].replace("&nfy=1", "")
        request_id = redirect_url.split("id=")[-1]
        await self.session.get(f"{BING_URL}{redirect_url}")
        polling_url = f"{BING_URL}/images/create/async/results/{request_id}?q={url_encoded_prompt}"
        # Poll for results
        if not self.quiet:
            print("Waiting for results...")
        time_sec = 0
        while True:
            if not self.quiet:
                print(".", end="", flush=True)
            # By default, timeout is 300s, change as needed
            response = await self.session.get(polling_url)
            if response.status_code != 200:
                raise NoResultsFound("Could not get results")
            content = response.text
            if content and content.find("errorMessage") == -1:
                break

            await asyncio.sleep(1)
            time_sec = time_sec + 1
            if time_sec >= max_generate_time_sec:
                raise TimeoutError("Out of generate time")
        # Use regex to search for src=""
        image_links = regex.findall(r'src="([^"]+)"', content)
        # Remove size limit
        normal_image_links = [link.split("?w=")[0] for link in image_links]
        # Remove duplicates
        normal_image_links = list(set(normal_image_links))

        # Bad images
        bad_images = [
            "https://r.bing.com/rp/in-2zU3AJUdkgFe7ZKv19yPBHVs.png",
            "https://r.bing.com/rp/TX9QuO3WzcCJz1uaaSwQAz39Kb0.jpg",
        ]
        for im in normal_image_links:
            if im in bad_images:
                raise NoResultsFound("Bad images")
        # No images
        if not normal_image_links:
            raise NoResultsFound("No images")
        return normal_image_links


async def async_image_gen(
        prompt: str,
        u_cookie=None,
        quiet=False,
        all_cookies=None,
):
    async with ImageGenAsync(
            u_cookie,
            quiet=quiet,
            all_cookies=all_cookies,
    ) as image_generator:
        return await image_generator.get_images(prompt)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-U", help="Auth cookie from browser", type=str)
    parser.add_argument("--cookie-file", help="File containing auth cookie", type=str)
    parser.add_argument(
        "--prompt",
        help="Prompt to generate images for",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--output-dir",
        help="Output directory",
        type=str,
        default="./output",
    )

    parser.add_argument(
        "--download-count",
        help="Number of images to download, value must be less than five",
        type=int,
        default=4,
    )

    parser.add_argument(
        "--debug-file",
        help="Path to the file where debug information will be written.",
        type=str,
    )

    parser.add_argument(
        "--quiet",
        help="Disable pipeline messages",
        action="store_true",
    )
    parser.add_argument(
        "--asyncio",
        help="Run ImageGen using asyncio",
        action="store_true",
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="Print the version number",
    )

    args = parser.parse_args()

    if args.version:
        sys.exit()

    # Load auth cookie
    cookie_json = None
    if args.cookie_file is not None:
        with contextlib.suppress(Exception):
            with open(args.cookie_file, encoding="utf-8") as file:
                cookie_json = json.load(file)

    if args.U is None and args.cookie_file is None:
        raise AuthCookieError("Could not find auth cookie")

    if args.download_count > 4:
        raise LimitExceeded("The number of downloads must be less than five")

    if not args.asyncio:
        # Create image generator
        image_generator = ImageGen(
            auth_cookie=args.U,
            quiet=args.quiet,
            all_cookies=cookie_json,
        )
        image_generator.save_images(
            image_generator.get_images(args.prompt),
            output_dir=args.output_dir,
            download_count=args.download_count,
        )
    else:
        asyncio.run(
            async_image_gen(
                prompt=args.prompt,
                u_cookie=args.U,
                quiet=args.quiet,
                all_cookies=cookie_json,
            ),
        )


if __name__ == "__main__":
    main()



from re_edge_gpt.image.image_genearation import ImageGen
from re_edge_gpt.image.image_genearation import ImageGenAsync
from re_edge_gpt.chat.re_edge_gpt import Chatbot
from re_edge_gpt.chat.re_edge_gpt import ConversationStyle

__all__ = [
    "Chatbot", "ConversationStyle", "ImageGen", "ImageGenAsync"
]



from PySide6.QtCore import QTimer
from PySide6.QtGui import QFontDatabase, Qt
from PySide6.QtWidgets import QWidget, QGridLayout, QPushButton, QScrollArea, QComboBox, QLabel, \
    QPlainTextEdit, QLineEdit, QBoxLayout, QCheckBox

from re_edge_gpt.ui.chat.input_dialog import ChatInputDialog
from re_edge_gpt.ui.chat.chat_thread import ChatThread, DELEGATE_CHAT, PANEL_MESSAGE_QUEUE
from re_edge_gpt.ui.chat.speech_to_text import ChatSpeechToText


class ChatWidget(QWidget):

    def __init__(self):
        super().__init__()
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.voice_input = None
        self.grid_layout = QGridLayout()
        self.grid_layout.setContentsMargins(0, 0, 0, 0)
        # Init
        self.chat_input = ChatInputDialog()
        self.choose_style_combobox = QComboBox()
        self.choose_style_combobox.addItems(["creative", "precise", "balanced"])
        self.choose_style_combobox.currentTextChanged.connect(self.change_style)
        # New topic button
        self.new_topic_button = QPushButton("New topic")
        self.new_topic_button.clicked.connect(self.new_topic)
        # Start button
        self.start_button = QPushButton("Start chat")
        self.start_button.clicked.connect(self.start_chat)
        # Chat panel
        self.chat_panel = QPlainTextEdit()
        self.chat_panel.setLineWrapMode(self.chat_panel.LineWrapMode.NoWrap)
        self.chat_panel.setReadOnly(True)
        self.chat_panel_scroll_area = QScrollArea()
        self.chat_panel_scroll_area.setWidgetResizable(True)
        self.chat_panel_scroll_area.setViewportMargins(0, 0, 0, 0)
        self.chat_panel_scroll_area.setWidget(self.chat_panel)
        self.chat_panel.setFont(QFontDatabase.font(self.font().family(), "", 16))
        # Font size combobox
        self.font_size_label = QLabel("Font size")
        self.font_size_combobox = QComboBox()
        for font_size in range(2, 101, 2):
            self.font_size_combobox.addItem(str(font_size))
        self.font_size_combobox.setCurrentText("16")
        self.font_size_combobox.currentTextChanged.connect(self.update_panel_text_size)
        # Close delay combobox
        self.close_delay_label = QLabel("Toast close delay")
        self.close_delay_combobox = QComboBox()
        for sec in range(1, 101, 1):
            self.close_delay_combobox.addItem(str(sec))
        self.close_delay_combobox.setCurrentText("10")
        # Locale box
        self.locale_label = QLabel("Country code")
        self.locale_input = QLineEdit()
        self.locale_input.setText("zh-tw")
        self.local_box = QBoxLayout(QBoxLayout.Direction.LeftToRight)
        self.local_box.addWidget(self.locale_label)
        self.local_box.addWidget(self.locale_input)
        # Start voice input
        self.start_voice_input_button = QPushButton("Start voice input")
        self.start_voice_input_button.clicked.connect(self.start_voice_input)
        # Show toast checkbox
        self.show_toast_checkbox_label = QLabel("Show toast?")
        self.show_toast_checkbox = QCheckBox()
        # Add to layout
        self.grid_layout.addWidget(self.choose_style_combobox, 0, 0)
        self.grid_layout.addWidget(self.close_delay_label, 0, 1)
        self.grid_layout.addWidget(self.close_delay_combobox, 0, 2)
        self.grid_layout.addWidget(self.font_size_label, 0, 3)
        self.grid_layout.addWidget(self.font_size_combobox, 0, 4)
        self.grid_layout.addLayout(self.local_box, 0, 5)
        self.grid_layout.addWidget(self.show_toast_checkbox_label, 0, 6)
        self.grid_layout.addWidget(self.show_toast_checkbox, 0, 7)
        self.grid_layout.addWidget(self.new_topic_button, 0, 8)
        self.grid_layout.addWidget(self.start_voice_input_button, 0, 9)
        self.grid_layout.addWidget(self.start_button, 0, 10)
        self.grid_layout.addWidget(self.chat_panel_scroll_area, 1, 0, -1, -1)
        # update panel timer
        self.update_panel_timer = QTimer()
        self.update_panel_timer.setInterval(10)
        self.update_panel_timer.timeout.connect(self.update_panel)
        self.update_panel_timer.start()
        self.setLayout(self.grid_layout)

    def update_panel(self):
        if not PANEL_MESSAGE_QUEUE.empty():
            text = PANEL_MESSAGE_QUEUE.get_nowait()
            self.chat_panel.appendPlainText(text)
            self.chat_panel.appendPlainText("\n")

    def update_panel_text_size(self):
        self.chat_panel.setFont(
            QFontDatabase.font(self.font().family(), "", int(self.font_size_combobox.currentText())))

    def start_chat(self) -> None:
        self.chat_input = ChatInputDialog(
            close_time=int(self.close_delay_combobox.currentText()) * 1000,
            font_size=int(self.font_size_combobox.currentText()),
            show_toast=self.show_toast_checkbox.isChecked()
        )
        self.chat_input.show()
        self.chat_input.send_text_button.clicked.connect(self.send_chat)

    def start_voice_input(self):
        self.voice_input = ChatSpeechToText()
        self.voice_input.show()
        self.voice_input.send_text_button.clicked.connect(self.send_voice_chat)

    def send_voice_chat(self):
        chat_thread = ChatThread(self.voice_input.voice_text_edit.text(), self.locale_input.text())
        chat_thread.start()

    def send_chat(self):
        chat_thread = ChatThread(self.chat_input.chat_input.toPlainText(), self.locale_input.text())
        chat_thread.start()

    def change_style(self):
        DELEGATE_CHAT.change_style(self.choose_style_combobox.currentText())

    def new_topic(self):
        DELEGATE_CHAT.new_topic(self.chat_panel)
        self.chat_input.close()

    def close_chat_ui(self):
        if self.chat_input.isVisible():
            self.chat_input.close()
        self.chat_input = None


    def close(self) -> bool:
        self.close_chat_ui()
        return super().close()



# Error messages
error_timeout = "Your request has timed out."
error_redirect = "Redirect failed"
error_blocked_prompt = (
    "Your prompt has been blocked by Bing. Try to change any bad words and try again."
)
error_being_reviewed_prompt = "Your prompt is being reviewed by Bing. Try to change any sensitive words and try again."
error_noresults = "Could not get results"
error_unsupported_lang = "\nthis language is currently not supported by bing"
error_bad_images = "Bad images"
error_no_images = "No images"
error_image_create_failed = "Image create failed pls check cookie or old image still creating"
# Action messages
sending_message = "Sending request..."
wait_message = "Waiting for results..."
download_message = "\nDownloading images..."
add_plugin_failed_message = "Can't add plugin, something wrong"



import asyncio
import json
from pathlib import Path
from queue import Queue
from threading import Thread

from re_edge_gpt import Chatbot, ConversationStyle
from PySide6.QtWidgets import QPlainTextEdit


class DelegateChat(object):

    def __init__(self):
        self.chat_bot = None
        self.style = ConversationStyle.creative

    def new_topic(self, message_panel: QPlainTextEdit):
        self.chat_bot = None
        message_panel.clear()

    def change_style(self, style: str):
        if style == "creative":
            self.style = ConversationStyle.creative
        elif style == "precise":
            self.style = ConversationStyle.precise
        else:
            self.style = ConversationStyle.balanced


class ChatThread(Thread):

    def __init__(self, chat_send_message: str, locale: str):
        super().__init__()
        self.current_message = None
        self.chat_send_message = chat_send_message
        PANEL_MESSAGE_QUEUE.put_nowait("Q: " + chat_send_message)
        self.locale = locale
        if DELEGATE_CHAT.chat_bot is not None:
            self.chat_bot = DELEGATE_CHAT.chat_bot

    def run(self) -> None:
        try:
            chat_response = dict()

            async def send_chat_async():
                nonlocal chat_response
                if DELEGATE_CHAT.chat_bot is None:
                    cookies = json.loads(open(
                        str(Path(str(Path.cwd()) + "/bing_cookies.json")), encoding="utf-8").read())
                    bot = await Chatbot.create(cookies=cookies)
                    response = await bot.ask(
                        prompt=self.chat_send_message, conversation_style=DELEGATE_CHAT.style, locale=self.locale
                    )
                    chat_response = response
                    DELEGATE_CHAT.chat_bot = bot
                else:
                    response = await DELEGATE_CHAT.chat_bot.ask(
                        prompt=self.chat_send_message, conversation_style=DELEGATE_CHAT.style)
                    chat_response = response

            asyncio.run(send_chat_async())
            self.current_message = chat_response
            if self.current_message is not None:
                for text_dict in self.current_message.get("item").get("messages"):
                    if text_dict.get("author") == "bot":
                        response_text: str = text_dict.get("text")
                        if response_text is not None and not response_text.isspace():
                            MESSAGE_QUEUE.put_nowait(response_text)
                            PANEL_MESSAGE_QUEUE.put_nowait("A: " + response_text)
        except Exception as error:
            EXCEPTION_QUEUE.put_nowait(repr(error))
            raise error


MESSAGE_QUEUE = Queue()
PANEL_MESSAGE_QUEUE = Queue()
DELEGATE_CHAT = DelegateChat()
EXCEPTION_QUEUE = Queue()


import json

from flask import Blueprint, request, jsonify

from re_edge_gpt import Chatbot, ConversationStyle, ImageGenAsync

re_edge_gpt_blueprint_instance = Blueprint(
    "re_edge_gpt_blueprint", __name__, url_prefix="/ReEdgeGPT",
    static_folder="static", template_folder="templates")

bot = {

}


async def setup_chatbot():
    cookies = json.loads(open("bing_cookies.json", encoding="utf-8").read())
    chatbot = await Chatbot.create(cookies=cookies)
    bot.update({"chatbot": chatbot})
    return chatbot


async def setup_imagebot():
    auth_cooker = open("bing_cookies.txt", "r+").read()
    imagebot = ImageGenAsync(auth_cookie=auth_cooker)
    bot.update({"imagebot": imagebot})
    return imagebot


@re_edge_gpt_blueprint_instance.route("/chat", methods=["POST"])
async def chat():
    """
    Send a prompt and chat mode to chat with Bing
    ---
   post:
    description: Chat with bing
    tags:
      - GPT
    requestBody:
      description: Data that chat needed
      content:
        application/json:
          schema:
            properties:
              prompt:
                type: string
                example: How to boil the eggs
              conversation_style:
                type: string
                enum: ["creative", "balanced", "precise"]
                default: balanced
              simplify_response:
                type: boolean
              attachment:
                type: object
                properties:
                  image_url:
                    type: string
                    default: null
                  filename:
                    type: string
                    default: null
                  base64_image:
                    type: string
                    default: null
    responses:
      200:
        description: Success chat
        content:
          application/json:
            schema:
              properties:
                text:
                  type: string
                author:
                  type: string
                sources:
                  type: string
                sources_link:
                  type: string
                suggestions:
                  type: array
                  example: [
                    "How do I know if the eggs are fresh?",
                    "What is a soft-boiled egg?",
                    "Can you give me some recipes with boiled eggs?"
                    ]
                messages_left:
                  type: integer
                  example: 29
                max_messages:
                  type: integer
                  example: 30
    """
    prompt = request.get_json()["prompt"]
    conversation_style = request.get_json()["conversation_style"]
    conversation_style = {
        "creative": ConversationStyle.creative,
        "balanced": ConversationStyle.balanced,
        "precise": ConversationStyle.precise,
    }.get(conversation_style, ConversationStyle.balanced)
    simplify_response = request.get_json()["simplify_response"]
    attachment = request.get_json()["attachment"]
    if simplify_response is None:
        simplify_response = True
    if bot.get("chatbot") is None:
        chatbot = await setup_chatbot()
    else:
        chatbot = bot.get("chatbot")
    try:
        response = await chatbot.ask(
            prompt=prompt,
            conversation_style=conversation_style,
            simplify_response=simplify_response,
            attachment=attachment
        )
        return jsonify(response)
    except Exception as error:
        raise error
    finally:
        if chatbot is not None:
            await chatbot.close()


@re_edge_gpt_blueprint_instance.route("/image", methods=["POST"])
async def image():
    """
    Send a prompt to generate image
    ---
    post:
      description: Generate image using DALL3-E
      tags:
        - GPT
      requestBody:
        description: prompt
        content:
          application/json:
            schema:
              properties:
                prompt:
                  type: string
      responses:
        200:
          description: Success generate image
          content:
            application/json:
              schema:
                properties:
                  images:
                    type: array
                    example: [
                      "image_url1",
                      "image_url2",
                      "image_url3"
                      ]
    """
    prompt = request.get_json()["prompt"]
    if bot.get("chatbot") is None:
        imagebot = await setup_imagebot()
    else:
        imagebot = bot.get("imagebot")
    image_list = await imagebot.get_images(prompt)
    return jsonify({"images": image_list})



import uuid
from datetime import datetime
from typing import Union

from re_edge_gpt.chat.conversation_style import CONVERSATION_STYLE_TYPE
from re_edge_gpt.chat.conversation_style import ConversationStyle
from re_edge_gpt.utils.utilities import get_location_hint_from_locale
from re_edge_gpt.utils.utilities import get_ran_hex
from re_edge_gpt.utils.utilities import guess_locale


class ChatHubRequest:
    def __init__(
            self,
            conversation_signature: str,
            client_id: str,
            conversation_id: str,
            invocation_id: int = 3,
    ) -> None:
        self.struct: dict = {}

        self.client_id: str = client_id
        self.conversation_id: str = conversation_id
        self.conversation_signature: str = conversation_signature
        self.invocation_id: int = invocation_id

    def update(
            self,
            prompt: str,
            conversation_style: CONVERSATION_STYLE_TYPE,
            webpage_context: Union[str, None] = None,
            search_result: bool = False,
            locale: str = guess_locale(),
            image_url: str = None,
            plugins: list = None,
            message_type: str = "Chat"
    ) -> None:
        if conversation_style:
            if not isinstance(conversation_style, ConversationStyle):
                conversation_style = getattr(ConversationStyle, conversation_style)
        message_id = str(uuid.uuid4())
        request_id = str(uuid.uuid4())
        # Get the current local time
        now_local = datetime.now()

        # Get the current UTC time
        now_utc = datetime.utcnow()

        # Calculate the time difference between local and UTC time
        timezone_offset = now_local - now_utc

        # Get the offset in hours and minutes
        offset_hours = int(timezone_offset.total_seconds() // 3600)
        offset_minutes = int((timezone_offset.total_seconds() % 3600) // 60)

        # Format the offset as a string
        offset_string = f"{offset_hours:+03d}:{offset_minutes:02d}"

        # Get current time
        timestamp = datetime.now().strftime("%Y-%m-%dT%H:%M:%S") + offset_string
        style = conversation_style.name.capitalize()  # Make first letter uppercase
        not_in_style = {"creative_classic": "CreativeClassic"}
        self.struct = {
            "arguments": [
                {
                    "optionsSets": conversation_style.value,
                    "allowedMessageTypes": [
                        "ActionRequest",
                        "Chat",
                        "ConfirmationCard",
                        "Context",
                        "InternalSearchQuery",
                        "InternalSearchResult",
                        "Disengaged",
                        "InternalLoaderMessage",
                        "Progress",
                        "RenderCardRequest",
                        "RenderContentRequest",
                        "AdsQuery",
                        "SemanticSerp",
                        "GenerateContentQuery",
                        "SearchQuery",
                        "GeneratedCode",
                        "InternalTasksMessage"
                    ],
                    "sliceIds": [
                        "disbotgrtcf",
                        "ntbkgold2",
                        "ntbkf1",
                        "qna10",
                        "thdnsrch",
                        "slangcf",
                        "vnextr100",
                        "vnext100",
                        "vnextvoice",
                        "rdlidncf",
                        "semserpnomlbg",
                        "semserpnoml",
                        "srchqryfix",
                        "cacntjndcae",
                        "edgenorrwrap",
                        "cmcpupsalltf",
                        "sunoupsell",
                        "313dynaplfs0",
                        "0312hrthrots0",
                        "0317immslotsc",
                        "228pyfiles0",
                        "kcclickthrucf",
                        "sportsatis0",
                        "0317dc1pro",
                        "defgrey",
                        "ssadsv4chtiidnoifbm",
                        "adsltmdsc",
                        "ssadsv2nocm"
                    ],
                    "plugins": plugins,
                    "isStartOfSession": self.invocation_id == 3,
                    "message": {
                        "locale": locale,
                        "market": locale,
                        "region": locale[-2:],  # en-US -> US
                        "locationHints": get_location_hint_from_locale(locale),
                        "timestamp": timestamp,
                        "author": "user",
                        "inputMethod": "Keyboard",
                        "text": prompt,
                        "messageType": message_type,
                        "messageId": message_id,
                        "requestId": request_id,
                        "imageUrl": image_url if image_url else None,
                        "originalImageUrl": image_url if image_url else None,
                    },
                    "tone": style if style not in not_in_style.keys()
                    else not_in_style.get(style),
                    "requestId": request_id,
                    "conversationSignature": self.conversation_signature,
                    "participant": {
                        "id": self.client_id,
                    },
                    "conversationId": self.conversation_id,
                },
            ],
            "invocationId": str(self.invocation_id),
            "target": "chat",
            "type": 4,
        }
        if search_result:
            have_search_result = [
                "InternalSearchQuery",
                "InternalSearchResult",
                "InternalLoaderMessage",
                "RenderCardRequest",
            ]
            self.struct["arguments"][0]["allowedMessageTypes"] += have_search_result
        if webpage_context:
            self.struct["arguments"][0]["previousMessages"] = [
                {
                    "author": "user",
                    "description": webpage_context,
                    "contextType": "WebPage",
                    "messageType": "Context",
                    "messageId": "discover-web--page-ping-mriduna-----",
                },
            ]
        self.invocation_id += 1

        # print(timestamp)



class NotAllowedToAccess(Exception):
    pass


class UnSupportLanguage(Exception):
    pass


class PromptBlocked(Exception):
    pass


class ImageCreateFailed(Exception):
    pass


class NoResultsFound(Exception):
    pass


class AuthCookieError(Exception):
    pass


class LimitExceeded(Exception):
    pass


class InappropriateContentType(Exception):
    pass


class ResponseError(Exception):
    pass


class PluginError(Exception):
    pass



import queue
import sys
import time

from PySide6.QtCore import QTimer, Qt
from PySide6.QtWidgets import QWidget, QPushButton, QBoxLayout, QLineEdit
from speech_recognition import Microphone
from speech_recognition import Recognizer
from speech_recognition import RequestError, UnknownValueError
from threading import Thread

LISTENER_QUEUE = queue.Queue()


def callback(recognizer: Recognizer, audio):
    try:
        text = recognizer.recognize_google(audio)
        LISTENER_QUEUE.put_nowait(text)
    except (RequestError, UnknownValueError) as error:
        print(repr(error), file=sys.stderr)


class ChatSpeechToText(QWidget):

    def __init__(self):
        super().__init__()
        # UI
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.box_layout = QBoxLayout(QBoxLayout.Direction.LeftToRight)
        self.voice_text_edit = QLineEdit()
        self.start_listen_button = QPushButton("Start voice input")
        self.start_listen_button.clicked.connect(self.start_listener_thread)
        self.send_text_button = QPushButton("Send prompt")
        self.box_layout.addWidget(self.voice_text_edit)
        self.box_layout.addWidget(self.start_listen_button)
        self.box_layout.addWidget(self.send_text_button)
        self.setLayout(self.box_layout)
        # Listener Timer
        self.listener_timer = QTimer()
        self.listener_timer.setInterval(100)
        self.listener_timer.timeout.connect(self.update_voice_edit)
        self.listener_timer.start()

    def start_listener_thread(self):
        listener_thread = Thread(target=self.start_listener)
        listener_thread.daemon = True
        listener_thread.start()

    @classmethod
    def start_listener(cls):
        recognizer = Recognizer()
        microphone = Microphone()
        with microphone as source:
            recognizer.adjust_for_ambient_noise(source, duration=0.1)
        stop_listening = recognizer.listen_in_background(microphone, callback)
        for receive_sound_time in range(50):
            time.sleep(0.1)
        stop_listening(wait_for_stop=False)

    def update_voice_edit(self):
        if not LISTENER_QUEUE.empty():
            self.voice_text_edit.setText(str(LISTENER_QUEUE.get_nowait()))

    def close(self) -> bool:
        self.listener_timer.stop()
        return super().close()


from enum import Enum

try:
    from typing import Literal, Union
except ImportError:
    from typing_extensions import Literal
from typing import Optional


class ConversationStyle(Enum):
    creative = [
        "nlu_direct_response_filter",
        "deepleo",
        "disable_emoji_spoken_text",
        "responsible_ai_policy_235",
        "enablemm",
        "dv3sugg",
        "autosave",
        "iyxapbing",
        "iycapbing",
        "h3imaginative",
        "clgalileo",
        "gencontentv3",
        "uquopt",
        "sunoupsell",
        "gndlogcf",
        "flxvsearch",
        "noknowimg",
        "eredirecturl"
    ]
    creative_classic = [
        "nlu_direct_response_filter",
        "deepleo",
        "disable_emoji_spoken_text",
        "responsible_ai_policy_235",
        "enablemm",
        "dv3sugg",
        "autosave",
        "iyxapbing",
        "iycapbing",
        "h3imaginative",
        "clgalileo",
        "gencontentv3",
        "uquopt",
        "sunoupsell",
        "gndlogcf",
        "flxvsearch",
        "noknowimg",
        "eredirecturl"
    ]
    balanced = [
        "nlu_direct_response_filter",
        "deepleo",
        "disable_emoji_spoken_text",
        "responsible_ai_policy_235",
        "enablemm",
        "dv3sugg",
        "autosave",
        "iyxapbing",
        "iycapbing",
        "enable_user_consent",
        "fluxmemcst",
        "galileo",
        "saharagenconv5",
        "dc1mncp",
        "uquopt",
        "sunoupsell",
        "crkt2t",
        "immslots",
        "cpproname",
        "vidtoppb",
        "gptv1desc2",
        "eredirecturl"
    ]
    precise = [
        "nlu_direct_response_filter",
        "deepleo",
        "disable_emoji_spoken_text",
        "responsible_ai_policy_235",
        "enablemm",
        "dv3sugg",
        "autosave",
        "iyxapbing",
        "iycapbing",
        "enable_user_consent",
        "fluxmemcst",
        "h3precise",
        "clgalileo",
        "uquopt",
        "sunoupsell",
        "crkt2t",
        "flxvsearchans",
        "noknowimg",
        "eredirecturl"
    ]


CONVERSATION_STYLE_TYPE = Optional[
    Union[ConversationStyle, Literal["creative", "balanced", "precise"]]
]



"""
Main.py
"""
from __future__ import annotations

from .chathub import *
from .constants import APP_ID, PLUGINS
from .conversation import *
from .request import *
from ..utils.exception.exception_message import add_plugin_failed_message
from ..utils.exception.exceptions import LimitExceeded, PluginError


class Chatbot:
    """
    Combines everything to make it seamless
    """

    def __init__(
            self,
            proxy: str | None = None,
            cookies: list[dict] | None = None,
    ) -> None:
        self.proxy: str | None = proxy
        self.chat_hub: ChatHub = ChatHub(
            Conversation(self.proxy, cookies=cookies),
            proxy=self.proxy,
            cookies=cookies,
        )

    async def add_plugins(self, cookies: list[dict], mode: str, conversation_id: str,
                          plugin_ids: Union[None, List[str]] = None) -> None:
        cookies_dict = {}
        if cookies is not None:
            for cookie in cookies:
                cookies_dict[cookie["name"]] = cookie["value"]

        async with httpx.AsyncClient(
                proxies=self.proxy,
                timeout=30,
                headers=HEADERS_INIT_CONVER,
                cookies=cookies_dict
        ) as client:
            if plugin_ids is not None:
                for plugin_id in plugin_ids:
                    if mode == "Bing":
                        response = await client.post(
                            f"https://www.bing.com/codex/plugins/conversation/add"
                            f"?conversationId={conversation_id}"
                            f"&appid={APP_ID}&pluginId={plugin_id}"
                        )
                    else:
                        response = await client.post(
                            f"https://copilot.microsoft.com/turing/plugins/conversation/add"
                            f"?conversationId={conversation_id}"
                            f"&appid={APP_ID}&pluginId={plugin_id}"
                        )
                    if response.status_code != 200:
                        raise PluginError(add_plugin_failed_message)

    @staticmethod
    async def create(
            proxy: str | None = None,
            cookies: list[dict] | None = None,
            mode: str = "Bing",
            plugin_ids: Union[None, List[str]] = None
    ) -> Chatbot:
        self = Chatbot.__new__(Chatbot)
        self.proxy = proxy
        self.mode = mode
        conversation = await Conversation.create(self.proxy, cookies=cookies, mode=mode)
        if plugin_ids is not None:
            new_plugin_list = [PLUGINS.get(plugin, plugin) for plugin in plugin_ids]
            await self.add_plugins(cookies, mode, conversation.struct["conversationId"], new_plugin_list)
        self.chat_hub = ChatHub(
            conversation,
            proxy=self.proxy,
            cookies=cookies,
            mode=mode
        )
        return self

    async def ask(
            self,
            prompt: str,
            wss_link: str = "wss://sydney.bing.com/sydney/ChatHub",
            conversation_style: CONVERSATION_STYLE_TYPE = None,
            webpage_context: str | None = None,
            search_result: bool = False,
            locale: str = guess_locale(),
            simplify_response: bool = False,
            attachment: dict[str, str] = None,
            remove_options: list = None,
            add_options: list = None,
            plugins: list = None,
            message_type: str = "Chat"
    ):
        """
        Ask a question to the bot
        :param prompt: The prompt to ask Bing
        :param wss_link: The link to the Bing web service
        :param conversation_style: The style of the Bing chat
        :param webpage_context: U don't need use this param in normal use
        :param search_result: Search web True or False
        :param locale: Bing service locale
        :param simplify_response: Simplify response True or False
        :param attachment: Send image
            attachment example:
                For url using
                attachment={"image_url": r"<image_url>"})
                For local file using
                attachment={"filename": r"<file_path>"})
                For base64 image using
                attachment={"base64_image": r"<base64_image_str>"})
        :param remove_options remove options from Style
        :param add_options add options to Style
        :param plugins: plugin name as list of strings
        :param message_type: Bing chat's message type
        """
        async for final, response in self.chat_hub.ask_stream(
                prompt=prompt,
                conversation_style=conversation_style,
                wss_link=wss_link,
                webpage_context=webpage_context,
                search_result=search_result,
                locale=locale,
                attachment=attachment,
                remove_options=remove_options,
                add_options=add_options,
                plugins=plugins,
                message_type=message_type,
        ):
            if final:
                if not simplify_response:
                    return response
                messages_left = (response.get("item").get("throttling").get("maxNumUserMessagesInConversation")
                                 - response.get("item").get("throttling").get(
                            "numUserMessagesInConversation",
                            0,
                        ))
                if messages_left == 0:
                    raise LimitExceeded("Max messages reached")
                message = {}
                return_data = {}
                for msg in reversed(response.get("item").get("messages")):
                    if msg.get("author") == "bot":
                        old_message = message.get("text")
                        if old_message:
                            old_message = old_message + " \n "
                        else:
                            old_message = ""
                        message.update({
                            "author": "bot",
                            "text": old_message + msg.get("text", "")
                        })
                        if msg.get("hiddenText") is not None:
                            hidden_text = msg.get("hiddenText")
                            hidden_text = hidden_text.replace("RequestId=", "")
                            return_data.update({"requestId": hidden_text})
                if not message:
                    raise NoResultsFound("No message found")
                image_create_text = ""
                suggestions = []
                source_keys = []
                source_values = []
                for detail in reversed(response.get("item").get("messages")):
                    suggestion_responses = detail.get("suggestedResponses", {})
                    source_attr = detail.get("sourceAttributions", {})
                    if suggestion_responses:
                        for suggestion in suggestion_responses:
                            suggestions.append(suggestion.get("text"))
                    if source_attr:
                        if isinstance(source_attr, dict):
                            for key, value in source_attr.items():
                                source_keys.append(key)
                                source_values.append(value)
                        if isinstance(source_attr, list):
                            for source_dict in source_attr:
                                if isinstance(source_dict, dict):
                                    source_keys.append(source_dict.get("providerDisplayName", ""))
                                    source_values.append(source_dict.get("seeMoreUrl", ""))
                    if detail.get("contentType") == "IMAGE" and detail.get("messageType") == "GenerateContentQuery":
                        image_create_text = detail.get("text")
                return_data.update({
                    "text": message["text"],
                    "author": message["author"],
                    "source_keys": source_keys,
                    "source_values": source_values,
                    "suggestions": suggestions,
                    "image_create_text": image_create_text,
                    "messages_left": messages_left,
                    "max_messages": response["item"]["throttling"][
                        "maxNumUserMessagesInConversation"
                    ],
                    "messageId": response.get("item").get("messages")[0]["messageId"],
                })
                return return_data
        return {}

    async def ask_stream(
            self,
            prompt: str,
            wss_link: str = "wss://sydney.bing.com/sydney/ChatHub",
            conversation_style: CONVERSATION_STYLE_TYPE = None,
            raw: bool = False,
            webpage_context: str | None = None,
            search_result: bool = False,
            locale: str = guess_locale(),
            remove_options: list = None,
            add_options: list = None
    ) -> Generator[bool, dict | str, None]:
        """
        Ask a question to the bot
        """
        async for response in self.chat_hub.ask_stream(
                prompt=prompt,
                conversation_style=conversation_style,
                wss_link=wss_link,
                raw=raw,
                webpage_context=webpage_context,
                search_result=search_result,
                locale=locale,
                remove_options=remove_options,
                add_options=add_options
        ):
            yield response

    async def close(self) -> None:
        """
        Close the connection
        """
        await self.chat_hub.close()

    async def reset(self) -> None:
        """
        Reset the conversation
        """
        await self.close()
        self.chat_hub = ChatHub(
            await Conversation.create(self.proxy, cookies=self.chat_hub.cookies),
            proxy=self.proxy,
            cookies=self.chat_hub.cookies,
        )



from apispec import APISpec
from apispec_webframeworks.flask import FlaskPlugin
from flask import Flask, Blueprint, render_template


def write_swagger_doc(openapi_yaml_content: str):
    with open("openapi.yaml", "w+") as openapi_file:
        openapi_file.write(openapi_yaml_content)


def bind_swagger(flask_app: Flask):
    from re_edge_gpt.api.re_edge_gpt_blueprint import chat
    from re_edge_gpt.api.re_edge_gpt_blueprint import image

    flask_api_docs_view = [
        chat, image
    ]

    spec = APISpec(
        title="ReEdgeGPT Example API Doc",
        version="1.0.0",
        openapi_version="3.0.2",
        servers=[{"url": "http://localhost:8888"}],
        plugins=[FlaskPlugin()]
    )

    with flask_app.app_context():
        for doc_view in flask_api_docs_view:
            spec.path(view=doc_view)

    write_swagger_doc(spec.to_yaml())



from flask import Flask, render_template
from flask_cors import CORS

from re_edge_gpt.api.re_edge_gpt_blueprint import re_edge_gpt_blueprint_instance
from re_edge_gpt.api.swagger_config import bind_swagger


def create_app() -> Flask:
    flask_app = Flask(__name__)
    blueprints = [re_edge_gpt_blueprint_instance]
    for blueprint in blueprints:
        flask_app.register_blueprint(blueprint)

    @flask_app.route("/", methods=["GET"])
    async def index():
        return render_template("index.html")

    CORS(flask_app)
    bind_swagger(flask_app)

    return flask_app



import base64
import json

import aiohttp

from re_edge_gpt.chat.constants import IMAGE_HEADER

payload = {
    "imageInfo": {},
    "knowledgeRequest": {
        "invokedSkills": ["ImageById"],
        "subscriptionId": "Bing.Chat.Multimodal",
        "invokedSkillsRequestData": {"enableFaceBlur": True},
        "convoData": {
            "convoid": "",
            "convotone": "Balanced"
        }
    }
}


async def upload_image_url(image_url: str, conversation_id: str, cookies: dict,
                           proxy: str = None, face_blur: bool = True):
    async with aiohttp.ClientSession(
            headers=IMAGE_HEADER, cookies=cookies
    ) as session:
        url = "https://www.bing.com/images/kblob"

        new_payload = payload
        new_payload.get("knowledgeRequest").update(
            {"invokedSkillsRequestData": {"enableFaceBlur": face_blur}})
        new_payload.get("imageInfo").update({"url": image_url})
        new_payload.get("knowledgeRequest").get("convoData").update({"convoid": conversation_id})
        data = aiohttp.FormData()
        data.add_field('knowledgeRequest', json.dumps(new_payload), content_type="application/json")
        async with session.post(url, data=data, proxy=proxy) as resp:
            if resp.status != 200:
                raise ConnectionError("Upload image failed")
            return (await resp.json())["blobId"]


async def upload_image(conversation_id: str, cookies: dict, filename: str = None,
                       base64_image: str = None, proxy: str = None, face_blur: bool = True):
    async with aiohttp.ClientSession(
            headers=IMAGE_HEADER, cookies=cookies
    ) as session:
        url = "https://www.bing.com/images/kblob"

        new_payload = payload
        new_payload.get("knowledgeRequest").update(
            {"invokedSkillsRequestData": {"enableFaceBlur": face_blur}})
        new_payload.get("knowledgeRequest").get("convoData").update({"convoid": conversation_id})

        if filename is not None:
            with open(filename, 'rb') as f:
                file_data = f.read()
                image_base64 = base64.b64encode(file_data)
        elif base64_image is not None:
            image_base64 = base64_image
        else:
            raise ConnectionRefusedError('no image provided')

        data = aiohttp.FormData()
        data.add_field('knowledgeRequest', json.dumps(new_payload), content_type="application/json")
        data.add_field('imageBase64', image_base64, content_type="application/octet-stream")
        async with session.post(url, data=data, proxy=proxy) as resp:
            if resp.status != 200:
                raise ConnectionError("Upload image failed")
            return (await resp.json())["blobId"]



import asyncio
import json
import ssl
import sys
from time import time
from typing import Generator
from typing import List
from typing import Union
from urllib import parse

import aiohttp
import certifi
import httpx

from re_edge_gpt.chat.constants import DELIMITER, SYDNEY_INIT_HEADER, SYDNEY_HEADER
from re_edge_gpt.chat.constants import HEADERS
from re_edge_gpt.chat.constants import HEADERS_INIT_CONVER
from re_edge_gpt.chat.conversation_style import CONVERSATION_STYLE_TYPE
from re_edge_gpt.chat.proxy import get_proxy
from re_edge_gpt.image.upload_image import upload_image, upload_image_url
from re_edge_gpt.utils.utilities import append_identifier
from re_edge_gpt.utils.utilities import guess_locale
from .conversation import Conversation
from .request import ChatHubRequest
from ..utils.exception.exceptions import NoResultsFound, ResponseError

ssl_context = ssl.create_default_context()
ssl_context.load_verify_locations(certifi.where())


async def _initial_handshake(wss) -> None:
    await wss.send_str(append_identifier({"protocol": "json", "version": 1}))
    await wss.receive_str()
    await wss.send_str(append_identifier({"type": 6}))


class ChatHub:
    def __init__(
            self,
            conversation: Conversation,
            proxy: str = None,
            cookies: Union[List[dict], None] = None,
            mode: str = "Bing",
    ) -> None:
        self.aio_session = None
        self.request: ChatHubRequest
        self.loop: bool
        self.task: asyncio.Task
        self.mode = mode
        self.conversation_id = conversation.struct["conversationId"]
        self.cookies = cookies
        self.proxy: str = get_proxy(proxy)
        self.request = ChatHubRequest(
            conversation_signature=conversation.struct["conversationSignature"],
            client_id=conversation.struct["clientId"],
            conversation_id=conversation.struct["conversationId"],
        )
        if self.conversation_id is None:
            self.conversation_id = self.request.conversation_id
        if self.mode == "Bing":
            header = HEADERS_INIT_CONVER
        else:
            header = SYDNEY_INIT_HEADER
        self.session = httpx.AsyncClient(
            proxies=self.proxy,
            timeout=900,
            headers=header,
        )
        if conversation.struct.get("encryptedConversationSignature"):
            self.encrypted_conversation_signature = conversation.struct["encryptedConversationSignature"]
        else:
            self.encrypted_conversation_signature = None
        self.conversation = conversation

    async def ask_stream(
            self,
            prompt: str,
            wss_link: str = None,
            conversation_style: CONVERSATION_STYLE_TYPE = None,
            raw: bool = False,
            webpage_context: Union[str, None] = None,
            search_result: bool = False,
            locale: str = guess_locale(),
            # Use for attachment
            attachment: dict = None,
            remove_options: list = None,
            add_options: list = None,
            plugins: list = None,
            message_type: str = "Chat"
    ) -> Generator[bool, Union[dict, str], None]:
        """ """
        if self.encrypted_conversation_signature is not None:
            wss_link = wss_link or "wss://sydney.bing.com/sydney/ChatHub"
            wss_link += f"?sec_access_token={parse.quote(self.encrypted_conversation_signature)}"
        cookies = {}
        if self.cookies is not None:
            for cookie in self.cookies:
                cookies[cookie["name"]] = cookie["value"]
        self.aio_session = aiohttp.ClientSession(cookies=cookies)
        if self.mode == "Bing":
            header = HEADERS
        else:
            header = SYDNEY_HEADER
        # Check if websocket is closed
        wss = await self.aio_session.ws_connect(
            wss_link or "wss://sydney.bing.com/sydney/ChatHub",
            ssl=ssl_context,
            headers=header,
            proxy=self.proxy,
        )
        await _initial_handshake(wss)
        # Image
        image_url = None
        if attachment is not None:
            if attachment.get("image_url") is not None:
                response = await upload_image_url(
                    **attachment, conversation_id=self.conversation_id, cookies=cookies)
            else:
                response = await upload_image(
                    **attachment, conversation_id=self.conversation_id, cookies=cookies)
            if response:
                image_url = f"https://www.bing.com/images/blob?bcid={response}"
        # Construct a ChatHub request
        if remove_options is not None:
            for option in remove_options:
                if option in conversation_style.value:
                    conversation_style.value.remove(option)
        if add_options is not None:
            for option in add_options:
                conversation_style.value.append(option)
        self.request.update(
            prompt=prompt,
            conversation_style=conversation_style,
            webpage_context=webpage_context,
            search_result=search_result,
            locale=locale,
            image_url=image_url,
            plugins=plugins,
            message_type=message_type,
        )
        # Send request
        await wss.send_str(append_identifier(self.request.struct))
        resp_txt = ""
        result_text = ""
        resp_txt_no_link = ""
        retry_count = 5
        while not wss.closed:
            msg = await wss.receive_str()
            if not msg:
                retry_count -= 1
                if retry_count == 0:
                    raise NoResultsFound("No response from server")
                continue
            if isinstance(msg, str):
                objects = msg.split(DELIMITER)
            else:
                continue
            for obj in objects:
                if int(time()) % 6 == 0:
                    await wss.send_str(append_identifier({"type": 6}))
                if obj is None or not obj:
                    continue
                response = json.loads(obj)
                if response.get("type") == 1 and response["arguments"][0].get("messages"):
                    if (response["arguments"][0]["messages"][0]["contentOrigin"] != "Apology") and not raw:
                        try:
                            resp_txt = result_text + response["arguments"][0][
                                "messages"
                            ][0]["adaptiveCards"][0]["body"][0].get("text", "")
                            resp_txt_no_link = result_text + response["arguments"][0][
                                "messages"
                            ][0].get("text", "")
                            if response["arguments"][0]["messages"][0].get(
                                    "messageType",
                            ):
                                resp_txt = (
                                        resp_txt
                                        + response["arguments"][0]["messages"][0][
                                            "adaptiveCards"
                                        ][0]["body"][0]["inlines"][0].get("text")
                                        + "\n"
                                )
                                result_text = (
                                        result_text
                                        + response["arguments"][0]["messages"][0][
                                            "adaptiveCards"
                                        ][0]["body"][0]["inlines"][0].get("text")
                                        + "\n"
                                )
                        except KeyError:
                            pass
                    if not raw:
                        yield False, resp_txt
                elif response.get("type") == 2:
                    if response["item"]["result"].get("error"):
                        await self.close()
                        raise ResponseError(
                            f"{response['item']['result']['value']}: {response['item']['result']['message']} \n"
                            f"Full exception: {response}",
                        )
                    if response["item"]["messages"][-1]["contentOrigin"] == "Apology" and resp_txt:
                        response["item"]["messages"][-1]["text"] = resp_txt_no_link
                        response["item"]["messages"][-1]["adaptiveCards"][0]["body"][0][
                            "text"
                        ] = resp_txt
                        print(
                            "Preserved the message from being deleted",
                            file=sys.stderr,
                        )
                    await wss.close()
                    if not self.aio_session.closed:
                        await self.aio_session.close()
                    yield True, response
                    return
                if response.get("type") != 2:
                    if response.get("type") == 6:
                        await wss.send_str(append_identifier({"type": 6}))
                    elif response.get("type") == 7:
                        await wss.send_str(append_identifier({"type": 7}))
                    elif raw:
                        yield False, response

    async def close(self) -> None:
        await self.session.aclose()

    async def get_conversation(self) -> dict:
        return {
            "conversationId": self.conversation.struct["conversationId"],
            "clientId": self.conversation.struct["clientId"],
            "encryptedConversationSignature": self.conversation.struct["encryptedConversationSignature"],
            "conversationSignature": self.conversation.struct["conversationSignature"],
        }

    async def set_conversation(self, conversation_dict: dict) -> None:
        self.conversation.struct["conversationId"] = conversation_dict.get("conversationId")
        self.conversation.struct["clientId"] = conversation_dict.get("clientId")
        self.conversation.struct[
            "encryptedConversationSignature"] = conversation_dict.get("encryptedConversationSignature")
        self.conversation.struct["conversationSignature"] = conversation_dict.get("conversationSignature")

    async def delete_conversation(self, conversation_id: str = None, client_id: str = None,
                                  encrypted_conversation_signature: str = None, conversation_signature: str = None
                                  ) -> None:
        self.conversation.struct["conversationId"] = conversation_id
        self.conversation.struct["clientId"] = client_id
        self.conversation.struct[
            "encryptedConversationSignature"] = encrypted_conversation_signature
        self.conversation.struct["conversationSignature"] = conversation_signature
        delete_conversation_url = "https://sydney.bing.com/sydney/DeleteSingleConversation"
        await self.session.post(
            delete_conversation_url,
            json={
                "conversationId": conversation_id,
                "conversationSignature": conversation_signature,
                "encrypted_conversation_signature": encrypted_conversation_signature,
                "participant": {"id": client_id},
                "source": "cib",
                "optionsSets": ["autosave"],
            },
        )



import json
import locale
import random
import sys
from typing import Union

from re_edge_gpt.chat.constants import DELIMITER
from re_edge_gpt.chat.locale import LocationHint


def append_identifier(msg: dict) -> str:
    # Convert dict to json string
    return json.dumps(msg, ensure_ascii=False) + DELIMITER


def get_ran_hex(length: int = 32) -> str:
    return "".join(random.choice("0123456789abcdef") for _ in range(length))


def get_location_hint_from_locale(locale_str: str) -> Union[dict, None]:
    locale_str = locale_str.lower()
    if locale_str == "en-gb":
        hint = LocationHint.UK.value
    elif locale_str == "en-ie":
        hint = LocationHint.EU.value
    elif locale_str == "zh-cn":
        hint = LocationHint.CHINA.value
    else:
        hint = LocationHint.USA.value
    return hint.get("LocationHint")


def guess_locale() -> str:
    if sys.platform.startswith("win"):
        return "en-us"
    loc, _ = locale.getlocale()
    return loc.replace("_", "-") if loc else "en-us"



import sys

from PySide6.QtWidgets import QMainWindow, QApplication
from qt_material import QtStyleTools, apply_stylesheet

from re_edge_gpt.ui.chat.main_widget import ChatWidget


class ChatMainUI(QMainWindow, QtStyleTools):

    def __init__(self):
        super().__init__()
        self.central_widget = ChatWidget()
        self.setCentralWidget(self.central_widget)


def start_chat_ui() -> None:
    main_app = QApplication(sys.argv)
    window = ChatMainUI()
    apply_stylesheet(main_app, theme='dark_amber.xml')
    window.showMaximized()
    sys.exit(main_app.exec())



from PySide6.QtCore import Qt, QRect, QTimer
from PySide6.QtGui import QPainter, QFontDatabase, QColor
from PySide6.QtWidgets import QWidget


class ChatToast(QWidget):

    def __init__(self, text: str, close_time: int = 10000, font_size: int = 16):
        super().__init__()
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setWindowFlag(
            Qt.WindowType.WindowTransparentForInput |
            Qt.WindowType.FramelessWindowHint |
            Qt.WindowType.Tool |
            Qt.WindowType.WindowStaysOnTopHint
        )
        self.text = text
        self.font_size = font_size
        self.draw_font = QFontDatabase.font(self.font().family(), "", self.font_size)
        self.draw_font.setBold(True)
        self.close_timer = QTimer()
        self.close_timer.setInterval(close_time)
        self.close_timer.timeout.connect(self.close)
        self.close_timer.start()

    def paintEvent(self, event) -> None:
        painter = QPainter(self)
        painter.save()
        painter.setFont(
            self.draw_font
        )
        painter.setPen(QColor(0, 255, 0))
        painter.drawText(
            QRect(self.x(), self.y(), self.width(), self.height()),
            Qt.AlignmentFlag.AlignCenter | Qt.TextFlag.TextWordWrap,
            self.text
        )
        if not self.close_timer.isActive():
            self.close_timer.start()
        painter.restore()

    def mousePressEvent(self, event) -> None:
        super().mousePressEvent(event)

    def mouseDoubleClickEvent(self, event) -> None:
        super().mouseDoubleClickEvent(event)

    def mouseGrabber(self) -> None:
        super().mouseGrabber()

    def close(self) -> bool:
        self.close_timer.stop()
        self.deleteLater()
        return super().close()
